pub mod app;
pub mod pages {
    pub mod edit_wall;
    pub mod routes;
    pub mod wall;
}
pub mod components {
    pub use button::Button;
    pub use header::StyledHeader;
    pub use problem::Problem;

    pub mod button;
    pub mod header;
    pub mod problem;
}
pub mod resources {
    use crate::codec::ron::Ron;
    use crate::codec::ron::RonEncoded;
    use crate::models::{self};
    use leptos::prelude::Get;
    use leptos::prelude::Signal;
    use leptos::server::Resource;
    use server_fn::ServerFnError;

    type RonResource<T> = Resource<Result<T, ServerFnError>, Ron>;

    pub fn wall_by_uid(wall_uid: Signal<models::WallUid>) -> RonResource<models::Wall> {
        Resource::new_with_options(
            move || wall_uid.get(),
            move |wall_uid| async move { crate::server_functions::get_wall(wall_uid).await.map(RonEncoded::into_inner) },
            false,
        )
    }

    pub fn problem_by_uid(wall_uid: Signal<models::WallUid>, problem_uid: Signal<models::ProblemUid>) -> RonResource<models::Problem> {
        Resource::new_with_options(
            move || (wall_uid.get(), problem_uid.get()),
            move |(wall_uid, problem_uid)| async move {
                crate::server_functions::get_problem(wall_uid, problem_uid)
                    .await
                    .map(RonEncoded::into_inner)
            },
            false,
        )
    }

    pub fn problems_for_wall(wall_uid: Signal<models::WallUid>) -> RonResource<Vec<models::Problem>> {
        Resource::new_with_options(
            move || wall_uid.get(),
            move |wall_uid| async move { crate::server_functions::get_problems_for_wall(wall_uid).await.map(RonEncoded::into_inner) },
            false,
        )
    }
}
pub mod codec;
pub mod models;
pub mod server_functions;

#[cfg(feature = "ssr")]
pub mod server;

#[cfg(feature = "hydrate")]
#[wasm_bindgen::prelude::wasm_bindgen]
pub fn hydrate() {
    use crate::app::*;

    console_error_panic_hook::set_once();

    tracing_subscriber::fmt()
        .with_env_filter(
            tracing_subscriber::EnvFilter::builder()
                .with_default_directive(tracing::level_filters::LevelFilter::DEBUG.into())
                .from_env_lossy(),
        )
        .with_writer(
            // To avoide trace events in the browser from showing their JS backtrace
            tracing_subscriber_wasm::MakeConsoleWriter::default().map_trace_level_to(tracing::Level::DEBUG),
        )
        // For some reason, if we don't do this in the browser, we get a runtime error.
        .without_time()
        .with_ansi(false)
        .init();

    leptos::mount::hydrate_body(App);
}
