use crate::components;
use crate::components::header::HeaderItem;
use crate::components::header::HeaderItems;
use crate::components::header::StyledHeader;
use crate::models;
use crate::models::WallUid;
use leptos::Params;
use leptos::prelude::*;
use leptos_router::params::Params;

#[derive(Params, PartialEq, Clone)]
struct RouteParams {
    // Is never None
    wall_uid: Option<models::WallUid>,
}

#[component]
#[tracing::instrument(skip_all)]
pub fn Routes() -> impl IntoView {
    tracing::debug!("Enter");

    let params = leptos_router::hooks::use_params::<RouteParams>();
    let wall_uid = Signal::derive(move || {
        params
            .get()
            .expect("gets wall_uid from URL")
            .wall_uid
            .expect("wall_uid param is never None")
    });

    let wall = crate::resources::wall_by_uid(wall_uid);
    let problems = crate::resources::problems_for_wall(wall_uid);

    let header_items = HeaderItems {
        left: vec![HeaderItem {
            text: "← Ascend".to_string(),
            link: Some("/".to_string()),
        }],
        middle: vec![HeaderItem {
            text: "ROUTES".to_string(),
            link: None,
        }],
        right: vec![],
    };

    let suspend = move || {
        Suspend::new(async move {
            let wall = wall.await;
            let problems = problems.await;

            let v = move || -> Result<_, ServerFnError> {
                let wall = wall.clone()?;
                let problems = problems.clone()?;

                let wall_dimensions = models::WallDimensions {
                    rows: wall.rows,
                    cols: wall.cols,
                };
                let problems_sample = move || problems.iter().take(10).cloned().collect::<Vec<_>>();

                Ok(view! {
                    <div>
                        <For
                            each=problems_sample
                            key=|problem| problem.uid
                            children=move |problem: models::Problem| {
                                view! { <Problem dim=wall_dimensions problem /> }
                            }
                        />
                    </div>
                })
            };

            view! { <ErrorBoundary fallback=|_errors| "error">{v}</ErrorBoundary> }
        })
    };

    view! {
        <div class="min-w-screen min-h-screen bg-slate-900">
            <StyledHeader items=header_items />

            <div class="container mx-auto mt-2">
                {move || view! { <Import wall_uid=wall_uid.get() /> }}
                <Suspense fallback=|| view! { <p>"loading"</p> }>{suspend}</Suspense>
            </div>
        </div>
    }
}

#[component]
#[tracing::instrument(skip_all)]
fn Problem(#[prop(into)] dim: Signal<models::WallDimensions>, #[prop(into)] problem: Signal<models::Problem>) -> impl IntoView {
    tracing::debug!("Enter");

    view! {
        <components::Problem dim problem />
        <p>{move || problem.get().name.clone()}</p>
    }
}

#[component]
#[tracing::instrument(skip_all)]
fn Import(wall_uid: WallUid) -> impl IntoView {
    tracing::debug!("ready");

    let import_from_mini_moonboard = Action::from(ServerAction::<ImportFromMiniMoonboard>::new());

    let onclick = move |_mouse_event| {
        import_from_mini_moonboard.dispatch(ImportFromMiniMoonboard { wall_uid });
    };

    view! {
        <p>"Import problems from"</p>
        <button on:click=onclick>"Mini Moonboard"</button>
    }
}

#[server(name = ImportFromMiniMoonboard)]
#[tracing::instrument]
async fn import_from_mini_moonboard(wall_uid: WallUid) -> Result<(), ServerFnError> {
    use crate::server::config::Config;
    use crate::server::db::Database;

    tracing::info!("Importing mini moonboard problems");

    let config = expect_context::<Config>();
    let db = expect_context::<Database>();

    crate::server::operations::import_mini_moonboard_problems(&config, db, wall_uid).await?;

    // TODO: Return information about what was done
    Ok(())
}
