use crate::codec::ron::RonEncoded;
use crate::components::button::Button;
use crate::components::header::HeaderItem;
use crate::components::header::HeaderItems;
use crate::components::header::StyledHeader;
use crate::models;
use crate::models::HoldRole;
use leptos::Params;
use leptos::prelude::*;
use leptos_router::params::Params;

#[derive(Params, PartialEq, Clone)]
struct RouteParams {
    wall_uid: Option<models::WallUid>,
}

#[component]
#[tracing::instrument(skip_all)]
pub fn Wall() -> impl IntoView {
    tracing::debug!("Enter");

    let params = leptos_router::hooks::use_params::<RouteParams>();
    let wall_uid = Signal::derive(move || {
        params
            .get()
            .expect("gets wall_uid from URL")
            .wall_uid
            .expect("wall_uid param is never None")
    });

    let wall = crate::resources::wall_by_uid(wall_uid);

    let problem_action = Action::new(move |&(wall_uid, problem_uid): &(models::WallUid, models::ProblemUid)| async move {
        tracing::info!("fetching");
        crate::server_functions::get_problem(wall_uid, problem_uid)
            .await
            .map(RonEncoded::into_inner)
    });
    let problem_signal = Signal::derive(move || {
        let v = problem_action.value().read_only().get();
        tracing::debug!("val: {:?}", v);
        v.and_then(Result::ok)
    });

    Effect::new(move |_prev_value| {
        problem_action.value().write_only().set(None);
        match wall.get() {
            Some(Ok(wall)) => {
                if let Some(problem_uid) = wall.random_problem() {
                    tracing::debug!("dispatching from effect");
                    problem_action.dispatch((wall.uid, problem_uid));
                }
            }
            Some(Err(_err)) => {}
            None => {}
        }
    });

    let header_items = move || HeaderItems {
        left: vec![],
        middle: vec![HeaderItem {
            text: "ASCEND".to_string(),
            link: None,
        }],
        right: vec![
            HeaderItem {
                text: "Routes".to_string(),
                link: Some(format!("/wall/{}/routes", wall_uid.get())),
            },
            HeaderItem {
                text: "Holds".to_string(),
                link: Some(format!("/wall/{}/edit", wall_uid.get())),
            },
        ],
    };

    leptos::view! {
        <div class="min-w-screen min-h-screen bg-slate-900">
            <StyledHeader items=Signal::derive(header_items) />

            <div class="m-2">
                <Suspense fallback=move || {
                    view! { <p>"Loading..."</p> }
                }>
                    {move || Suspend::new(async move {
                        tracing::info!("executing Suspend future");
                        let wall = wall.await?;
                        let v = view! {
                            <div class="grid grid-cols-[auto,1fr] gap-8">
                                <Grid wall=wall.clone() problem=problem_signal />

                                <div>
                                    <div>
                                        <p>
                                            {move || problem_signal.get().map(|p| p.name.clone())}
                                        </p>
                                        <p>
                                            {move || problem_signal.get().map(|p| p.set_by.clone())}
                                        </p>
                                    </div>

                                    <Button
                                        onclick=move |_| {
                                            if let Some(problem_uid) = wall.random_problem() {
                                                tracing::info!("dispatching from button click handler");
                                                problem_action.dispatch((wall.uid, problem_uid));
                                            }
                                        }
                                        text="➤ Next problem"
                                    />
                                </div>
                            </div>
                        };
                        Ok::<_, ServerFnError>(v)
                    })}
                </Suspense>
            </div>
        </div>
    }
}

#[component]
#[tracing::instrument(skip_all)]
fn Grid(wall: models::Wall, problem: Signal<Option<models::Problem>>) -> impl IntoView {
    tracing::debug!("Enter");

    let mut cells = vec![];
    for (&hold_position, hold) in &wall.holds {
        let role = move || problem.get().and_then(|p| p.holds.get(&hold_position).copied());
        let role = Signal::derive(role);
        let cell = view! { <Hold role hold=hold.clone() /> };
        cells.push(cell);
    }
    let grid_classes = format!("grid grid-rows-{} grid-cols-{} gap-3", wall.rows, wall.cols,);

    view! {
        <div class="grid grid-cols-[auto,1fr] gap-8">
            <div style="max-height: 90vh; max-width: 90vh;" class=move || { grid_classes.clone() }>
                {cells}
            </div>
        </div>
    }
}

#[component]
#[tracing::instrument(skip_all)]
fn Hold(hold: models::Hold, role: Signal<Option<HoldRole>>) -> impl IntoView {
    tracing::trace!("Enter");
    let class = move || {
        let role_classes = match role.get() {
            Some(HoldRole::Start) => Some("outline outline-offset-2 outline-green-500"),
            Some(HoldRole::Normal) => Some("outline outline-offset-2 outline-blue-500"),
            Some(HoldRole::Zone) => Some("outline outline-offset-2 outline-amber-500"),
            Some(HoldRole::End) => Some("outline outline-offset-2 outline-red-500"),
            None => Some("brightness-50"),
            // None => None,
        };
        let mut s = "bg-sky-100 aspect-square rounded".to_string();
        if let Some(c) = role_classes {
            s.push(' ');
            s.push_str(c);
        }
        s
    };

    let img = hold.image.map(|img| {
        let srcset = img.srcset();
        view! { <img class="object-cover w-full h-full" srcset=srcset /> }
    });

    tracing::trace!("view");
    view! { <div class=class>{img}</div> }
}
