use bincode::Bincode;
use redb::ReadTransaction;
use redb::TableDefinition;
use redb::WriteTransaction;
use serde::Deserialize;
use serde::Serialize;
use std::path::PathBuf;
use std::sync::Arc;

mod bincode;

const DB_FILE: &str = "datastore/private/ascend.redb";

#[derive(Debug, Clone)]
pub struct Database {
    db: Arc<redb::Database>,
}

impl Database {
    #[tracing::instrument(skip_all, err)]
    pub fn create() -> Result<Database, redb::Error> {
        let file = PathBuf::from(DB_FILE);

        // Create parent dirs
        if let Some(parent_dir) = file.parent() {
            std::fs::create_dir_all(parent_dir)?;
        }

        let db = redb::Database::create(file)?;
        Ok(Self { db: Arc::new(db) })
    }

    #[tracing::instrument(skip_all, err)]
    pub async fn read<T>(&self, f: impl FnOnce(&'_ ReadTransaction) -> Result<T, DatabaseOperationError>) -> Result<T, DatabaseOperationError> {
        tokio::task::block_in_place(|| {
            let dbtx = self.db.begin_read()?;
            f(&dbtx)
        })
    }

    #[tracing::instrument(skip_all, err)]
    pub async fn write<T>(&self, f: impl FnOnce(&'_ WriteTransaction) -> Result<T, DatabaseOperationError>) -> Result<T, DatabaseOperationError> {
        tokio::task::block_in_place(|| {
            let dbtx = self.db.begin_write()?;
            let res = f(&dbtx)?;
            dbtx.commit()?;
            Ok(res)
        })
    }

    #[tracing::instrument(skip_all)]
    pub async fn get_version(&self) -> Result<Option<Version>, DatabaseOperationError> {
        self.read(|dbtx| dbtx.open_table(TABLE_VERSION)?.get(()).map(|o| o.map(|v| v.value())).map_err(Into::into))
            .await
    }
}

#[derive(Debug, derive_more::Error, derive_more::Display, derive_more::From)]
#[display("DB operation error: {_variant}")]
pub enum DatabaseOperationError {
    #[display("redb error")]
    #[from(forward)]
    Redb(#[error(source)] redb::Error),

    #[from(ignore)]
    Custom(#[error(source)] Box<dyn std::error::Error + Send + Sync + 'static>),
}
impl DatabaseOperationError {
    pub fn custom(err: impl std::error::Error + Send + Sync + 'static) -> Self {
        Self::Custom(Box::new(err))
    }
}

pub const TABLE_VERSION: TableDefinition<(), Bincode<Version>> = TableDefinition::new("version");
#[derive(Serialize, Deserialize, Debug, derive_more::Display)]
#[display("{version}")]
pub struct Version {
    pub version: u64,
}
impl Version {
    pub fn current() -> Version {
        Version { version: current::VERSION }
    }
}

pub use v2 as current;

pub mod v2 {
    use crate::models;
    use crate::server::db::bincode::Bincode;
    use redb::TableDefinition;

    pub const VERSION: u64 = 2;

    pub const TABLE_ROOT: TableDefinition<(), Bincode<models::v2::Root>> = TableDefinition::new("root");
    pub const TABLE_WALLS: TableDefinition<Bincode<models::v2::WallUid>, Bincode<models::v2::Wall>> = TableDefinition::new("walls");
    pub const TABLE_PROBLEMS: TableDefinition<Bincode<(models::v2::WallUid, models::v2::ProblemUid)>, Bincode<models::v2::Problem>> =
        TableDefinition::new("problems");
}

pub mod v1 {
    use crate::models;
    use crate::server::db::bincode::Bincode;
    use redb::TableDefinition;

    pub const TABLE_ROOT: TableDefinition<(), Bincode<models::v1::PersistentState>> = TableDefinition::new("root");
}
