use crate::codec::ron::Ron;
use crate::codec::ron::RonEncoded;
use crate::models;
use leptos::server;
use server_fn::ServerFnError;

#[server(
    input = Ron,
    output = Ron,
    custom = RonEncoded
)]
#[tracing::instrument(skip_all, err(Debug))]
pub async fn get_walls() -> Result<RonEncoded<Vec<models::Wall>>, ServerFnError> {
    use crate::server::db::Database;
    use leptos::prelude::expect_context;
    use redb::ReadableTable;
    tracing::debug!("Enter");

    let db = expect_context::<Database>();

    let walls = db
        .read(|txn| {
            let walls_table = txn.open_table(crate::server::db::current::TABLE_WALLS)?;
            let walls: Vec<models::Wall> = walls_table.iter()?.map(|r| r.map(|(_, v)| v.value())).collect::<Result<_, _>>()?;
            Ok(walls)
        })
        .await?;

    tracing::debug!("Exit");

    Ok(RonEncoded::new(walls))
}

#[server(
    input = Ron,
    output = Ron,
    custom = RonEncoded
)]
#[tracing::instrument(skip_all, err(Debug))]
pub(crate) async fn get_wall(wall_uid: models::WallUid) -> Result<RonEncoded<models::Wall>, ServerFnError> {
    use crate::server::db::Database;
    use crate::server::db::DatabaseOperationError;
    use leptos::prelude::expect_context;
    tracing::debug!("Enter");

    #[derive(Debug, derive_more::Error, derive_more::Display)]
    enum Error {
        #[display("Wall not found: {_0:?}")]
        NotFound(#[error(not(source))] models::WallUid),
    }

    let db = expect_context::<Database>();

    let wall = db
        .read(|txn| {
            let walls_table = txn.open_table(crate::server::db::current::TABLE_WALLS)?;
            let wall = walls_table
                .get(wall_uid)?
                .ok_or(Error::NotFound(wall_uid))
                .map_err(DatabaseOperationError::custom)?
                .value();
            Ok(wall)
        })
        .await?;

    tracing::debug!("ok");

    Ok(RonEncoded::new(wall))
}

#[server(
    input = Ron,
    output = Ron,
    custom = RonEncoded
)]
#[tracing::instrument(err(Debug))]
pub(crate) async fn get_problems_for_wall(wall_uid: models::WallUid) -> Result<RonEncoded<Vec<models::Problem>>, ServerFnError> {
    use crate::server::db::Database;
    use crate::server::db::DatabaseOperationError;
    use leptos::prelude::expect_context;
    tracing::debug!("Enter");

    #[derive(Debug, derive_more::Error, derive_more::Display, derive_more::From)]
    enum Error {
        #[display("Wall not found: {_0:?}")]
        WallNotFound(#[error(not(source))] models::WallUid),

        DatabaseOperation(DatabaseOperationError),
    }

    async fn inner(wall_uid: models::WallUid) -> Result<Vec<models::Problem>, Error> {
        let db = expect_context::<Database>();

        let problems = db
            .read(|txn| {
                let walls_table = txn.open_table(crate::server::db::current::TABLE_WALLS)?;
                tracing::debug!("getting wall");
                let wall = walls_table
                    .get(wall_uid)?
                    .ok_or(Error::WallNotFound(wall_uid))
                    .map_err(DatabaseOperationError::custom)?
                    .value();
                tracing::debug!("got wall");
                drop(walls_table);

                tracing::debug!("open problems table");
                let problems_table = txn.open_table(crate::server::db::current::TABLE_PROBLEMS)?;
                tracing::debug!("opened problems table");

                let mut problems = Vec::new();
                for &problem_uid in &wall.problems {
                    if let Some(problem) = problems_table.get((wall_uid, problem_uid))? {
                        problems.push(problem.value());
                    }
                }
                Ok(problems)
            })
            .await?;

        Ok(problems)
    }

    let problems = inner(wall_uid).await.map_err(error_reporter::Report::new).map_err(ServerFnError::new)?;

    tracing::debug!("ok");

    Ok(RonEncoded::new(problems))
}

#[server(
    input = Ron,
    output = Ron,
    custom = RonEncoded
)]
#[tracing::instrument(skip_all, err(Debug))]
pub(crate) async fn get_problem(wall_uid: models::WallUid, problem_uid: models::ProblemUid) -> Result<RonEncoded<models::Problem>, ServerFnError> {
    use crate::server::db::Database;
    use crate::server::db::DatabaseOperationError;
    use leptos::prelude::expect_context;
    tracing::debug!("Enter");

    #[derive(Debug, derive_more::Error, derive_more::Display)]
    enum Error {
        #[display("Problem not found: {_0:?}")]
        NotFound(#[error(not(source))] models::ProblemUid),
    }

    let db = expect_context::<Database>();
    let problem = db
        .read(|txn| {
            let table = txn.open_table(crate::server::db::current::TABLE_PROBLEMS)?;
            let problem = table
                .get((wall_uid, problem_uid))?
                .ok_or(Error::NotFound(problem_uid))
                .map_err(DatabaseOperationError::custom)?
                .value();
            Ok(problem)
        })
        .await?;

    Ok(RonEncoded::new(problem))
}
